/** 
 * This is the public interface of InterComm.
 * @file intercomm.h
 * @author Christian Hansen (chansen@cs.umd.edu)
 *         Il-Chul Yoon (iyoon@cs.umd.edu)
 *         University of Maryland, College Park
 */

#ifndef INTERCOMM_H
#define INTERCOMM_H

#include "ezxml/ezxml.h"
/*#include "ic_config.h"*/

#define IC_BDECOMP 1
#define IC_TTABLE 2

/* Built-in types */

/** A program handle */
typedef struct {
  int nbNode;
  char* name;
  int* tid;
  int myPos;
  char* joined;
} IC_Program;

struct leaf_node;

/** A partial decomposition descriptor */
typedef struct decomp_TREE {
  int nDims;
  int num_of_partitions;
  struct decomp_TREE* children;
  int* keys;
  struct leaf_node* block_info;
} IC_Tree;

/** A decomposition descriptor */
typedef struct {
  int type;
  void* spec;
} IC_Desc;

typedef enum { R_Enum, R_HPF } T_RegionType;
typedef enum { R_replicated, R_distributed, R_local } T_RegionDistribution;
typedef enum { Mem_Copy, Mem_Pointer } T_MemOp;


/** An array region */
typedef struct {
  T_RegionType type;
  T_RegionDistribution distribution;
  T_MemOp mem_type;
  int size;
} IC_Region;

typedef struct OneSched_ {
  struct OneSched_ *next;
  int offset;
  int size;
} OneSched_s;

/** A communication schedule */
typedef struct {
  int nproc;
  int *size;
  OneSched_s **sched;
  OneSched_s **last;
} IC_Sched;


/**
 * An XJD component descriptor encapsulating an InterComm program.
 * Created dynamically based on XJD file during IC_Initialize.
 */
typedef struct IC_XJDComp_node {
    char* id;
    char* name;
    int ntask;
    
    IC_Program* comp;
    struct IC_XJDComp_node* next;
} IC_XJDComponent;

/**
 * An XJD region descriptor encapsulating an InterComm connection
 * between programs. Created dynamically based on XJD file during
 * IC_Initialize. 
 */
typedef struct IC_XJDRgn_node{
    char* myname;
    char* my_region_name;
    IC_Program* me;

    char* partnername;
    char* partner_region_name;
    IC_Program* partner;

    IC_Region** region;
    int   rgn_size;
    char* commtype; /* MxN , 1xN, or Nx1  */
    char* type;  /* data types supported by InterComm */
    int   tag;
    void* local_data;

    IC_Desc*  rgn_desc;
    IC_Sched* sched;
    
    struct IC_XJDRgn_node*  next;
} IC_XJDRegion;        


/**
 * An XJD descriptor. Contains the pointer to the head and tail
 * of the dynamically created XJD_Component and XJD_Region
 * list.
 */
typedef struct {
    char* version;

    IC_XJDComponent* cHeader; /* component header */
    IC_XJDComponent* cTail;
    IC_XJDRegion*    rHeader; /* region list header */
    IC_XJDRegion*    rTail;
    ezxml_t          xmlRoot;  
    
  /* Pointer to the data structures used for API implementing
     loosely-coupled message transfer */
    void* meou;
} IC_XJD;


/*
  An InterComm region handle. Contains the information for data transfer. 
typedef struct {
    IC_XJDRegion* rgn;
    
    tHandle thandle;  // for SML
     double timing;    // for SML
     iHandle ihandle;  // for SML
} IC_RegionHandle;
 */

/** Enumeration for column-major and row-major */
enum {
  IC_COLUMN_MAJOR = 1, /**< specifying column major */
  IC_ROW_MAJOR = 2, /**< specifying row major */
};


/**
 * Initialize the communication library
 * @param name   a unique program name
 * @param tasks  the number of tasks in this program
 * @param rank   the rank of this task
 * @return a handle for this program
 */
IC_Program* IC_Init(char* name, int tasks, int rank);

/**
 * Contact another program
 * @param name   the name of the other program (as given in its IC_Init call)
 * @param tasks  the number of tasks in the other program
 * @return a handle for the other program
 */
IC_Program* IC_Wait(char* name, int tasks);

/** 
 * Synchronizes two programs
 * @param myprog  a handle to this program
 * @param prog    a handle to the other program
 * @return status
 */
int IC_Sync(IC_Program* myprog, IC_Program* prog);

/**
 * Create a block decomposition
 * @param ndims   the number of dimensions
 * @param blocks  an array of block bound specifications
 * @param tasks   the task assignments for each block
 * @param count   the number of blocks
 * @param arrayOrder  specify whether it is row major or column major
 * @return the decomposition descriptor
 */
IC_Desc* IC_Create_bdecomp_desc(int ndims, int* blocks, int* tasks, int count, int arrayOrder);

/**
 * Create a translation table
 * @param globals  a list of global indices
 * @param locals   the local index for each global index
 * @param tasks    the task assignment for each global index
 * @param count    the number of global indices
 * @return the decomposition descriptor
 */
IC_Desc* IC_Create_ttable_desc(int* globals, int* locals, int* tasks, int count);

/**
 * Allocate a partial decomposition
 * @return the partial descriptor
 */
IC_Tree* IC_Create_bdecomp_tree();

/**
 * Create a partition for a partial descriptor
 * @param root the partial descriptor
 * @param dim the dimension to partition
 * @param count the number of partitions
 * @param indices the upper bounding index for each partition
 */
void IC_Section(IC_Tree* root, int dim, int count, int* indices);

/**
 * Create a partition for a partial descriptor
 * @param root the partial descriptor
 * @param dim the dimension to partition
 * @param block the block to partition
 * @param count the number of partitions
 * @param indices the upper bounding index for each partition
 * @return an array of partial descriptors which can then be partitioned further
 */
void IC_Partition(IC_Tree* root, int dim, int* block, int count, int* indices);

/**
 * Verify a partial descriptor and generate a complete descriptor
 * @param root the partial descriptor
 * @param ndims the number of dimensions
 * @param size the size of the global array
 * @param tasks the task assignments for each block
 * @param count the number of blocks
 * @param arrayOrder array order
 * @return the decomposition descriptor
 */
IC_Desc* IC_Verify_bdecomp_tree(IC_Tree* root, int ndims, int* size, int* tasks, int count, int arrayOrder);

/**
 * Allocate a block region structure
 * @param ndims   the number of dimensions
 * @param lower   the lower bounds
 * @param upper   the upper bounds
 * @param stride  the stride
 * @return the array region
 */
IC_Region* IC_Create_block_region(int ndims, int* lower, int* upper, int* stride);

/**
 * Allocate an enumeration region structure
 * @param indices  a list of global indices
 * @param count    the number of indices
 * @return the array region
 */
IC_Region* IC_Create_enum_region(int* indices, int count);


/**
 * Compute a schedule for the given regions
 * @param myprog      a handle to this program
 * @param prog        a handle to the other program
 * @param desc        a decomposition descriptor
 * @param region_set  a list of regions
 * @param set_size    the number of regions
 * @return the communication schedule
 */
IC_Sched* IC_Compute_schedule(IC_Program* myprog, IC_Program* prog, IC_Desc* desc, 
			      IC_Region** region_set, int set_size);

/* These TYPE functions are added as placeholders for Doxygen */
#define TYPE void

/**
 * Send a region to another program.  There is a send call for each C TYPE (char, 
 * short, int, long, float, double).
 * @li <b>int IC_Send_char(IC_Program* to, IC_Sched* sched, char* data, int tag)</b>
 * @li <b>int IC_Send_short(IC_Program* to, IC_Sched* sched, short* data, int tag)</b>
 * @li <b>int IC_Send_int(IC_Program* to, IC_Sched* sched, int* data, int tag)</b>
 * @li <b>int IC_Send_long(IC_Program* to, IC_Sched* sched, long* data, int tag)</b>
 * @li <b>int IC_Send_float(IC_Program* to, IC_Sched* sched, float* data, int tag)</b>
 * @li <b>int IC_Send_double(IC_Program* to, IC_Sched* sched, double* data, int tag)</b>
 * @param to     the receiving program
 * @param sched  the communication schedule
 * @param data   the local portion of the global array
 * @param tag    a message tag
 * @return status
 */
int IC_Send_TYPE(IC_Program* to, IC_Sched* sched, TYPE* data, int tag);

int IC_Send_char(IC_Program* to, IC_Sched* sched, char* data, int tag);
int IC_Send_short(IC_Program* to, IC_Sched* sched, short* data, int tag);
int IC_Send_int(IC_Program* to, IC_Sched* sched, int* data, int tag);
int IC_Send_long(IC_Program* to, IC_Sched* sched, long* data, int tag);
int IC_Send_float(IC_Program* to, IC_Sched* sched, float* data, int tag);
int IC_Send_double(IC_Program* to, IC_Sched* sched, double* data, int tag);

/** 
 * Receive a region from a program.  There is a receive call for each C TYPE (char,
 * short, int, long, float, double).
 * @li <b>int IC_Recv_char(IC_Program* from, IC_Sched* sched, char* data, int tag)</b>
 * @li <b>int IC_Recv_short(IC_Program* from, IC_Sched* sched, short* data, int tag)</b>
 * @li <b>int IC_Recv_int(IC_Program* from, IC_Sched* sched, int* data, int tag)</b>
 * @li <b>int IC_Recv_long(IC_Program* from, IC_Sched* sched, long* data, int tag)</b>
 * @li <b>int IC_Recv_float(IC_Program* from, IC_Sched* sched, float* data, int tag)</b>
 * @li <b>int IC_Recv_double(IC_Program* from, IC_Sched* sched, double* data, int tag)</b>
 * @param from   the sending program
 * @param sched  the communication schedule
 * @param data   the local portion of the global array
 * @param tag    a message tag
 * @return status
 */
int IC_Recv_TYPE(IC_Program* from, IC_Sched* sched, TYPE* data, int tag);

int IC_Recv_char(IC_Program* from, IC_Sched* sched, char* data, int tag);
int IC_Recv_short(IC_Program* from, IC_Sched* sched, short* data, int tag);
int IC_Recv_int(IC_Program* from, IC_Sched* sched, int* data, int tag);
int IC_Recv_float(IC_Program* from, IC_Sched* sched, float* data, int tag);
int IC_Recv_double(IC_Program* from, IC_Sched* sched, double* data, int tag);

/**
 * Frees a schedule
 * @param sched  a communication schedule 
 */
void IC_Free_sched(IC_Sched* sched);

/**
 * Frees a region
 * @param region  an array region
 */
void IC_Free_region(IC_Region* region);

/**
 * Frees a descriptor
 * @param desc  an array partitioning descriptor
 */
void IC_Free_desc(IC_Desc* desc);

/**
 * Fress a program handle
 * @param prog  a program handle
 */
void IC_Free_program(IC_Program* prog);

/**
 * Frees up internal data structures and
 * free the program handle for this program
 * @param myprog  the handle for this program
 */
void IC_Quit(IC_Program* myprog);

/**
 * Prints an error message for the last call that failed
 * @param msg  a message that will be prepended to the internal error message  
 */
void IC_Print_error(char* msg);


/**
 * This function initializes the underlying communication library and
 * creates the internal representation of the programs and regions
 * by parsing the given XML Job Description (XJD) file for further
 * communication between the programs. Basically, this function
 * encapsulates IC_Init, IC_Wait and IC_Sync.
 * @param pName  my program name
 * @param rank      my rank
 * @param xjdname   XML Job Description (XJD) file name
 * @param status     status, negative indicates an error
 * @return a pointer to the IC_XJD datatype, which contains or will contain the program and region list.
 */
IC_XJD* IC_Initialize(char* pName, int rank, char* xjdname, int* status);


/**
 * Create an InterComm region handle, which contains the information 
 * for data transfer. Created handle will have different information 
 * according to the given XJD version.
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param status     status, if negative, indicates an error
IC_RegionHandle* IC_InitHandle(IC_XJD* xjd, int* status);
 */


/**
 * Register detailed information on a set of regions (a set of array
 * blocks) into the internal program and region representation.
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param rgnset    an array of regions describing the data to be communicated 
 * @param set_size  the number of of regions in the array
 * @param set_name  name of the region set
 * @param desc      the InterComm global array descriptor
 * @param local_data the pointer to the local data array
 * @param status     status, if negative, indicates an error
 */
void IC_Register_region(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* status);


/**
 * Build communication schedule for all the regions in my program
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param status     status, if negative, indicates an error
 */
void IC_Commit_region(IC_XJD* xjd, int* status);


/**
 * Export a region. This function does not need the destination program
 * since it is described in the XJD file and the information becomes part
 * of internal representation of programs and regions during the
 * IC_Initialize. 
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param rgnset_name  name of the region set
 * @return status   
 */
int IC_Export(IC_XJD* xjd, char* rgnset_name);

    
/**
 * Import a region. This function does not need the destination program
 * since it is described in the XJD file and the information becomes part
 * of internal representation of programs and regions during the
 * IC_Initialize. 
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param rgnset_name  name of the region set
 * @return status  
 */
int IC_Import(IC_XJD* xjd, char* rgnset_name);


/**
 * Broadcast a local array into all processes in other program. There is a broadcast call for each C TYPE (char, short, int, float, double).
 * @li <b>int IC_BCast_Local_char(IC_Program* to, char* data, int nelems, int tag)</b>
 * @li <b>int IC_BCast_Local_short(IC_Program* to, short* data, int nelems, int tagt)</b>
 * @li <b>int IC_BCast_Local_int(IC_Program* to, int* data, int nelems, int tag)</b>
 * @li <b>int IC_BCast_Local_float(IC_Program* to, float* data, int nelems, int tag)</b>
 * @li <b>int IC_BCast_Local_double(IC_Program* to, double* data, int nelems, int tag)</b>
 * @param to     the receiving program
 * @param data   the array to broadcast
 * @param nelems the number of elemens in the data
 * @param tag    a message tag
 * @return status
 */
int IC_BCast_Local_TYPE(IC_Program* to, TYPE* data, int nelems, int tag);

int IC_BCast_Local_char(IC_Program* to, char* data, int nelems, int tag);
int IC_BCast_Local_short(IC_Program* to, short* data, int nelems, int tag);
int IC_BCast_Local_int(IC_Program* to, int* data, int nelems, int tag);
int IC_BCast_Local_float(IC_Program* to, float* data, int nelems, int tag);
int IC_BCast_Local_double(IC_Program* to, double* data, int nelems, int tag);


/**
 * Receive an array broadcasted from rank-0 process of other program. There is 
 * a receive call for each C TYPE (char, short, int, float, double).
 * @li <b>int IC_Recv_Local_char(IC_Program* from, char* data, int nelems, int tag)</b>
 * @li <b>int IC_Recv_Local_short(IC_Program* from, short* data, int nelems, int tag)</b>
 * @li <b>int IC_Recv_Local_int(IC_Program* from, int* data, int nelems, int tag)</b>
 * @li <b>int IC_Recv_Local_float(IC_Program* from, float* data, int nelems, int tag)</b>
 * @li <b>int IC_Recv_Local_double(IC_Program* from, double* data, int nelems, int tag)</b>
 * @param from   the receiving program
 * @param data   the array to broadcast
 * @param nelems the number of elemens in the data
 * @param tag    a message tag
 * @return status
 */
int IC_Recv_Local_TYPE(IC_Program* from, TYPE* data, int nelems, int tag);

int IC_Recv_Local_char(IC_Program* from, char* data, int nelems, int tag);
int IC_Recv_Local_short(IC_Program* from, short* data, int nelems, int tag);
int IC_Recv_Local_int(IC_Program* from, int* data, int nelems, int tag);
int IC_Recv_Local_float(IC_Program* from, float* data, int nelems, int tag);
int IC_Recv_Local_double(IC_Program* from, double* data, int nelems, int tag);



/**
 * 
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param rgnset_name  name of the region set
 * @param data      the data to broadcast
 * @param nelems    the data size, which is required since we cannot compute data size for local block
 * @return status     
 */
int IC_BCast_Local(IC_XJD* xjd, char* rgnset_name, void* data, int nelems);


/**
 * Receive an array broadcasted from rank-0 process of other program
 * 
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param rgnset_name  name of the region set
 * @param data      the data to broadcast
 * @param nelems    the data size, which is required since we cannot compute data size for local block
 * @return status     
 */
int IC_Recv_Local(IC_XJD* xjd, char* rgnset_name, void* data, int nelems);


/**
 * Release the allocated memory for the region handle
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param handle    handle to a region
 * @param icsts     status, if negative, indicates an error
void IC_Free_RegionHandle(IC_XJD* xjd, IC_RegionHandle* handle, int* icsts);
 */


/**
 * Release the allocated memory for internal representation of programs
 * and regions.
 * @param xjd       the InterComm application descriptor for programs and regions
 * @param icsts     status, if negative, indicates an error
 */
void IC_Finalize(IC_XJD* xjd, int* icsts);


#endif /* INTERCOMM_H */
