/**
 * This is the class for public high-level interface of InterComm.
 * @file IC_EndPoint.h
 * @author Henrique Andrade
 *         University of Maryland, College Park
 */

/* Henrique Andrade - March 2003
   University of Maryland, College Park

   Class Description:
  
   This is an abstract class for a MetaChaos endpoint. It implements the
   abstract methods described in the EndPoint class.*/

#ifndef IC_ENDPOINT_H
#define IC_ENDPOINT_H

#include "EndPoint.h"
#include <stdio.h>

#ifdef PPLUSPLUS
#include <A++.h>
#endif

#ifdef FORTRAN90
#include "carray_descriptor.h"
#endif

/*!
 This is the class for public high-level interface of InterComm.
*/
class IC_EndPoint : public EndPoint {
  public:

  /* the declarations below must be in sync with
     ../FortranInterface/intercomm_interface.h*/

  /** Enumeration for InterComm high-level for API errors.*/
    enum {
      IC_TAG          = 1,
      IC_COLUMN_MAJOR = 1,
      IC_ROW_MAJOR    = 2,
      IC_OK                         =   0,/**<no error*/
      IC_GENERIC_ERROR              =  -1,/**<generic error*/ 
      IC_INVALID_NDIM               =  -2,/**<invalid number of array dimensions*/
      IC_CANT_ALLOC_REGION          =  -3,/**<can't allocate InterComm array regions*/
      IC_CANT_GET_DA_DESCRIPTOR     =  -4,/**<can't obtain distributed array descriptor*/
      IC_CANT_COMPUTE_COMM_SCHEDULE =  -5,/**<can't compute communication schedule*/
      IC_COMM_FAILURE               =  -6,/**<communication (send/recv) failure */
      IC_INVALID_ENDPOINT_NAME      =  -7,/**<invalid endpoint name*/
      IC_INITIALIZATION_FAILURE     =  -8,/**<local program initialization failed*/
      IC_CANT_CONNECT_TO_REMOTE     =  -9,/**<can't connect to remote program*/
      IC_PVM_ERROR                  = -10,/**<PVM error*/
      IC_MPI_ERROR                  = -11,/**<MPI error*/
      IC_POINTER_TABLE_ERROR        = -12,/**<internal pointer translation table error*/
    };
    static const char * const IC_errors[];

/**
 * The constructor to create the communication endpoints
 * @param endpointName the name for the endpoint
 * @param mynproc the number of processors used by the local program
 * @param onproc the number of processors used by the remote program
 * @param myao the array ordering used by the local program
 * @param oao the array ordering used by the remote program
 * @param status the result of the endpoint creation operation 
*/
    /* to create a communication endpoint, one needs to name it (the format is
       "thisprogram'sname::otherprogram'sname"), define how many processors
       this program uses, as well as the other program, and define the array
       order used for this program as well as the other program*/
    IC_EndPoint(const char* endpointName, 
                const unsigned mynproc, const unsigned onproc, 
                const unsigned myao, const unsigned oao, 
                int& status);

    unsigned get_myprog_arrayorder(void);
    unsigned get_otherprog_arrayorder(void);   
    IC_Program* get_myprog(void);
    IC_Program* get_otherprog(void);   

    static void printErrorMessage(const char* msg, const int ic_err);

/**
 *  The destructor for the EndPoint. The application writer is supposed to invoke
 *  the delete C++ operator to ensure that the destructor properly finializes 
 *  InterComm 
*/

    virtual ~IC_EndPoint();
        
#ifdef PPLUSPLUS
  public:  
    /*Methods for High-Level API*/
    /* - compute schedule for each import/export*/
    virtual void exportArray(const intArray&, int, int&);
    virtual void importArray(const intArray&, int, int&);
    virtual void bcastLocalArray(const intArray&, int, int, int&);
    virtual void recvLocalArray(const intArray&, int, int, int&);
    virtual void exportArray(const floatArray&, int, int&);
    virtual void importArray(const floatArray&, int, int&);
    virtual void bcastLocalArray(const floatArray&, int, int, int&);
    virtual void recvLocalArray(const floatArray&, int, int, int&);
    virtual void exportArray(const doubleArray&, int, int&);
    virtual void importArray(const doubleArray&, int, int&);
    virtual void bcastLocalArray(const doubleArray&, int, int, int&);
    virtual void recvLocalArray(const doubleArray&, int, int, int&);
#endif

	
#ifdef FORTRAN90
  public:
    /* Methods for High-Level API*/
    virtual void exportArray(const F90_charArray&, int, int&);
    virtual void importArray(const F90_charArray&, int, int&);
    virtual void bcastLocalArray(const F90_charArray&, int, int, int&);
    virtual void recvLocalArray(const F90_charArray&, int, int, int&);
    virtual void exportArray(const F90_shortArray&, int, int&);
    virtual void importArray(const F90_shortArray&, int, int&);
    virtual void bcastLocalArray(const F90_shortArray&, int, int, int&);
    virtual void recvLocalArray(const F90_shortArray&, int, int, int&);
    virtual void exportArray(const F90_intArray&, int, int&);
    virtual void importArray(const F90_intArray&, int, int&);
    virtual void bcastLocalArray(const F90_intArray&, int, int, int&);
    virtual void recvLocalArray(const F90_intArray&, int, int, int&);
    virtual void exportArray(const F90_floatArray&, int, int&);
    virtual void importArray(const F90_floatArray&, int, int&);
    virtual void bcastLocalArray(const F90_floatArray&, int, int, int&);
    virtual void recvLocalArray(const F90_floatArray&, int, int, int&);
    virtual void exportArray(const F90_doubleArray&, int, int&);
    virtual void importArray(const F90_doubleArray&, int, int&);
    virtual void bcastLocalArray(const F90_doubleArray&, int, int, int&);
    virtual void recvLocalArray(const F90_doubleArray&, int, int, int&);
    
    static IC_Desc* buildDescriptor(const F90_charArray& a);
    static IC_Desc* buildDescriptor(const F90_shortArray& a);
    static IC_Desc* buildDescriptor(const F90_intArray& a);
    static IC_Desc* buildDescriptor(const F90_floatArray& a);
    static IC_Desc* buildDescriptor(const F90_doubleArray& a);
#endif

  protected:
  
    /* information needed by InterComm to carry out the communication
       pointers to the data structures describing the local and remote
       communication endpoints*/
    IC_Program *otherProgram;
    unsigned thisProgramAO, otherProgramAO;
    /* static stuff*/
    static IC_Program *thisProgram;
    static unsigned nendpoints;

};

#endif


