/**
 * $Author: normanlo $
 * $Date: 2006/05/04 18:06:07 $
 * $Revision: 1.1.2.135 $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */
 
#include <iostream>
#include <assert.h>
#include <string.h>
#include "IC_EndPointSet.h"


const char * const IC_EndPointSet::IC_epset_errors[] = {
    "no error",
    "generic error", 
    "invalid number of array dimensions",
    "can't allocate InterComm array regions",
    "can't obtain distributed array descriptor",
    "can't compute communication schedule",
    "communication (send/recv) failure",
    "invalid endpoint name",
    "this program's initialization failed",
    "can't connect to remote program",
    "PVM error. Is PVM running?",
    "MPI error",
    "internal pointer table error",
    "fail to initialize endpoint set from XJD",
    "fail to register array with array name",
    "fail to commit the arrays",
};

IC_EndPointSet::IC_EndPointSet(const char* xjdfile, const char* compname, int& stat) {

    std::cout << "arguments :" << XJDVERSION << " " << xjdfile << " " << compname << " " << stat << "\n";

    ezxml_t xmlRoot = read_xjd(xjdfile);
    assert(XJD_Validate(xmlRoot) == 0);

    this->myprogname = compname;
    this->xjdversion = strdup(get_version(xmlRoot));
    
    if(!strcmp(this->xjdversion, "1.5")){
        epset_init_1_5(xmlRoot, stat);
    }
    else if(!strcmp(this->xjdversion, "2.0")){
        epset_init_2_0(xmlRoot, stat);
    }
    assert(stat == 0);
      
}

IC_EndPointSet::~IC_EndPointSet(){
    map<string, IC_Sched*>::const_iterator p2s_iter; 
    map<string, IC_Region**>::const_iterator p2r_iter;
    map<string, IC_Desc*>::const_iterator p2d_iter;
    map<string, IC_EndPoint*>::const_iterator ep_iter;
  
    /* free schedule*/
    for (p2s_iter = this->port2sched.begin(); p2s_iter != this->port2sched.end(); ++p2s_iter) {
        string portname = p2s_iter->first;
        IC_Sched* sched = (IC_Sched*)p2s_iter->second;

        free(sched);
    }
 
    /* free region and regionset */
    for (p2r_iter = this->port2regionset.begin(); p2r_iter != this->port2regionset.end(); ++p2r_iter) {
        string portname = p2r_iter->first;
        IC_Region** rgnset = (IC_Region**)p2r_iter->second;

        /* because regionset has only one region in this case*/
        IC_Free_region(rgnset[0]);  
        free(rgnset);
    }
   
    /* free desc */
    for (p2d_iter = this->port2desc.begin(); p2d_iter != this->port2desc.end(); ++p2d_iter) {
        string portname = p2d_iter->first;
        IC_Desc* desc = (IC_Desc*)p2d_iter->second;

        /* because regionset has only one region in this case*/
        IC_Free_desc(desc);  
    }

    /* delete endpoint*/
    for (ep_iter = this->epmap.begin(); ep_iter != this->epmap.end(); ++ep_iter) {
        string epname = ep_iter->first;
        IC_EndPoint* ep = (IC_EndPoint*)ep_iter->second;

        delete ep;
    } 
        
    /* clear maps*/
    epmap.clear();
    port2ep.clear();
    port2desc.clear();
    port2regionset.clear();
    port2sched.clear();
    port2tag.clear();
   

#ifdef PPLUSPLUS
    /* clear arrays */
    ppp_int_arrmap.clear();
    ppp_float_arrmap.clear();
    ppp_double_arrmap.clear();
#endif

#ifdef FORTRAN90
    /* free the encapsulated arraydesc for each Fortran90 array */
    map<string, const F90_charArray*>::const_iterator f90_char_arrdesc_iter; 
    map<string, const F90_shortArray*>::const_iterator f90_short_arrdesc_iter; 
    map<string, const F90_intArray*>::const_iterator f90_int_arrdesc_iter; 
    map<string, const F90_floatArray*>::const_iterator f90_float_arrdesc_iter; 
    map<string, const F90_doubleArray*>::const_iterator f90_double_arrdesc_iter; 

    for (f90_char_arrdesc_iter = this->f90_char_arrmap.begin(); f90_char_arrdesc_iter != this->f90_char_arrmap.end(); ++f90_char_arrdesc_iter) {
        F90_charArray* a = (F90_charArray*)f90_char_arrdesc_iter->second;
        assert(a != NULL);
	delete a;
    }

    for (f90_short_arrdesc_iter = this->f90_short_arrmap.begin(); f90_short_arrdesc_iter != this->f90_short_arrmap.end(); ++f90_short_arrdesc_iter) {
        F90_shortArray* a = (F90_shortArray*)f90_short_arrdesc_iter->second;
        assert(a != NULL);
	delete a;
    }

    for (f90_int_arrdesc_iter = this->f90_int_arrmap.begin(); f90_int_arrdesc_iter != this->f90_int_arrmap.end(); ++f90_int_arrdesc_iter) {
        F90_intArray* a = (F90_intArray*)f90_int_arrdesc_iter->second;
        assert(a != NULL);
	delete a;
    }

    for (f90_float_arrdesc_iter = this->f90_float_arrmap.begin(); f90_float_arrdesc_iter != this->f90_float_arrmap.end(); ++f90_float_arrdesc_iter) {
        F90_floatArray* a = (F90_floatArray*)f90_float_arrdesc_iter->second;
        assert(a != NULL);
	delete a;
    }

    for (f90_double_arrdesc_iter = this->f90_double_arrmap.begin(); f90_double_arrdesc_iter != this->f90_double_arrmap.end(); ++f90_double_arrdesc_iter) {
        F90_doubleArray* a = (F90_doubleArray*)f90_double_arrdesc_iter->second;
        assert(a != NULL);
	delete a;
    }

    /* clear arrays	*/
    f90_char_arrmap.clear();
    f90_short_arrmap.clear();
    f90_int_arrmap.clear();
    f90_float_arrmap.clear();
    f90_double_arrmap.clear();
#endif
 
}

int IC_EndPointSet::XJD_Validate(ezxml_t xmlRoot){
  /* return -1 if error in validation, else return 0*/
    return 0;
}
    
int IC_EndPointSet::epset_init_1_5(ezxml_t xmlRoot, int& stat){
    stat = register_program(xmlRoot);
    std::cout << "Register Status of " << this->myprogname << " : " << stat << "\n";
    assert(stat == IC_OK);

    stat = inspect_connections(xmlRoot);
    assert(stat == IC_OK);

    return 0;
}

/**
 * Precondition: following class variables must be defined before.
 *   const char* myprogname;
 *   unsigned mynproc; 
 *   unsigned myao;
 */
int IC_EndPointSet::inspect_connections(ezxml_t xmlRoot){
    ezxml_t curConn, exp, imp, expport, impport, commtype, datatype;
    char* cid;
    char* pname;
    char* pname2;
    char* other_nproc;
    char* other_arrord;
    char* exportport;
    char* importport; 
    int tag = -1;
    
    unsigned onproc = 0;
    unsigned oao = 0;
    int stat = IC_OK;
    
    curConn = get_first_conn(xmlRoot);
    
    while(curConn){
        exp = ezxml_child(curConn, EXPORTER);
        pname = get_field_byID(exp->txt, xmlRoot, COMPNAME);
        
        if( !strcmp(this->myprogname, pname) ){
            fprintf(stderr, "\tappear as exporter: %s\n", pname);
            
            imp = ezxml_child(curConn, IMPORTER); /* get cid of partner program*/
            pname2 = get_field_byID(imp->txt, xmlRoot, COMPNAME);            

            other_nproc = get_field_byID(imp->txt, xmlRoot, N_NODE);
            if(other_nproc){
                onproc = (unsigned) atoi(other_nproc);
            }

            other_arrord = get_field_byID(imp->txt, xmlRoot, ARRAYORDER);
            if(other_arrord){
                if (!strcmp(other_arrord, "IC_COLUMN_MAJOR")){
                    oao = IC_EndPoint::IC_COLUMN_MAJOR;
                }
                else if(!strcmp(other_arrord, "IC_ROW_MAJOR")){
                    oao = IC_EndPoint::IC_ROW_MAJOR;
                }
                else{ /* error*/
                    fprintf(stderr, "Wrong array order in XJD.\n");
                    return IC_EPSET_INIT_FAILURE;
                }
            }

            expport = ezxml_child(curConn, EXPPORT);
            commtype = ezxml_child(curConn, COMMTYPE);
            datatype = ezxml_child(curConn, DATATYPE);
            tag = atoi(ezxml_child(curConn, TAG)->txt);
            
            assert( (pname) && (onproc > 0) && 
                    ((oao == IC_EndPoint::IC_COLUMN_MAJOR) || (oao == IC_EndPoint::IC_ROW_MAJOR)) &&
                    (expport != NULL) && (commtype != NULL) && (datatype != NULL)
                  );
                  
            addep(this->myprogname, this->mynproc, this->myao, 
                  pname2, onproc, oao, 
                  expport->txt, commtype->txt, datatype->txt, tag);
                
        }
        
        imp = ezxml_child(curConn, IMPORTER);
        pname = get_field_byID(imp->txt, xmlRoot, COMPNAME);
        
        if( !strcmp(this->myprogname, pname) ){
            fprintf(stderr, "\tappear as importer: %s\n", pname);                                    

            exp = ezxml_child(curConn, EXPORTER);
            pname2 = get_field_byID(exp->txt, xmlRoot, COMPNAME);            

            other_nproc = get_field_byID(exp->txt, xmlRoot, N_NODE);
            if(other_nproc){
                onproc = (unsigned) atoi(other_nproc);
            }

            other_arrord = get_field_byID(exp->txt, xmlRoot, ARRAYORDER);
            if(other_arrord){
                if (!strcmp(other_arrord, "IC_COLUMN_MAJOR")){
                    oao = IC_EndPoint::IC_COLUMN_MAJOR;
                }
                else if(!strcmp(other_arrord, "IC_ROW_MAJOR")){
                    oao = IC_EndPoint::IC_ROW_MAJOR;
                }
                else{ /* error*/
                    fprintf(stderr, "Wrong array order in XJD.\n");
                    return IC_EPSET_INIT_FAILURE;
                }
            }

            impport = ezxml_child(curConn, IMPPORT);
            commtype = ezxml_child(curConn, COMMTYPE);
            datatype = ezxml_child(curConn, DATATYPE);
            tag = atoi(ezxml_child(curConn, TAG)->txt);
            
            assert( (pname) && (onproc > 0) && 
                    ((oao == IC_EndPoint::IC_COLUMN_MAJOR) || (oao == IC_EndPoint::IC_ROW_MAJOR)) &&
                    (impport != NULL) && (commtype != NULL) && (datatype != NULL)
                  );
                  
            addep(this->myprogname, this->mynproc, this->myao,
                  pname2, onproc, oao, 
                  impport->txt, commtype->txt, datatype->txt, tag);
        }
        
        exp = imp = NULL;
        curConn = curConn->next;
    }
    
    return stat;
}


void IC_EndPointSet::addep(const char* myname, unsigned mynproc, unsigned myao,
                           char* othername, unsigned onproc, unsigned oao,
                           char* portname, char* commtype, char* datatype, int tag){
    int stat;                            
    string p1 = myname;
    string p2 = othername;
    string epname = p1 + ":" + p2;

    cout << endl;
    cout << "MyInfo: " << myname << " " << mynproc << " " << myao << endl;
    cout << "OtherInfo: " << othername << " " << onproc << " " << oao << endl;
    cout << endl;
    
    IC_EndPoint* ep = findep(epname);
    if(ep == NULL){
        cout << "\tRegistering new endpoint " << epname << " and port " << portname << "." << endl;

        ep = new IC_EndPoint(epname.c_str(), mynproc, onproc, myao, oao, stat);
        assert(stat == IC_EndPoint::IC_OK);
        
        this->epmap[epname] = ep;
        
        registerPort(ep, portname, commtype, datatype, tag, stat);
        assert(stat == IC_OK);
    }
    else{
        cout << "\tEndpoint exists! Just register port " << portname << "." << endl;

        registerPort(ep, portname, commtype, datatype, tag, stat);
        assert(stat == IC_OK);
    }

}

IC_EndPoint* IC_EndPointSet::findep(string epname){
    map<string, IC_EndPoint*>::const_iterator iter;

    iter = this->epmap.find(epname);
    if(iter == this->epmap.end()){
        #ifdef DEBUG
            cout << "no such endpoint name in endpoint map" << endl;
        #endif
        return NULL;
    }
    else{
        return (IC_EndPoint*)iter->second;
    }
}

/**
 * Assumption: portname is unique in a program (or a thread)
 */
void IC_EndPointSet::registerPort(IC_EndPoint* ep, char* ptname, 
                                  char* ctype, char* dtype, int tag, int& stat){
    string portname = ptname;

    this->port2ep[portname] = ep;
    this->port2tag[portname] = tag;
   
    /* 1xN port will not considered at commitArray 
       the portname for 1xN port appears in port2ep map and port2tag map. 
       But, not in port2desc map and not in real array maps
    */

    if(!strcmp(ctype, "MxN")){
      /* normal communication*/
      /* schedule must be generated between the processors.*/
        
        this->port2desc[portname] = NULL;

#ifdef PPLUSPLUS
        if(!strcmp(dtype, "int")){
            this->ppp_int_arrmap[portname] = NULL;
            cout << "\tintArray Map Size is : " << this->ppp_int_arrmap.size() << endl;
        }
        else if(!strcmp(dtype, "float")){
            this->ppp_float_arrmap[portname] = NULL;
            cout << "\tfloatArray Map Size is : " << this->ppp_float_arrmap.size() << endl;
        }
        else if(!strcmp(dtype, "double")){
            this->ppp_double_arrmap[portname] = NULL;
            cout << "\tdoubleArray Map Size is : " << this->ppp_double_arrmap.size() << endl;
        }
        else{
            cout << "\tInvalid datatype in XJD" << endl;
            stat = IC_EPSET_INIT_FAILURE;
            return;
        }
#endif

#ifdef FORTRAN90
        if(!strcmp(dtype, "int")){
            this->f90_int_arrmap[portname] = NULL;
            cout << "\tintArray Map Size is : " << this->f90_int_arrmap.size() << endl;
        }
        else if(!strcmp(dtype, "float")){
            this->f90_float_arrmap[portname] = NULL;
            cout << "\tfloatArray Map Size is : " << this->f90_float_arrmap.size() << endl;
        }
        else if(!strcmp(dtype, "double")){
            this->f90_double_arrmap[portname] = NULL;
            cout << "\tfloatArray Map Size is : " << this->f90_double_arrmap.size() << endl;
        }
        else{
            cout << "\tInvalid datatype in XJD" << endl;
            stat = IC_EPSET_INIT_FAILURE;
            return;
        }
#endif
    }
    
    stat = IC_OK;
}



int IC_EndPointSet::epset_init_2_0(ezxml_t xmlRoot, int& stat){
    return IC_OK;
}




ezxml_t IC_EndPointSet::read_xjd(const char* xjdname){
    FILE* fp; 
    
    if ((fp = fopen(xjdname, "r")) == NULL){
        fprintf(stderr, "Cannot open %s\n", xjdname);
        return NULL;
    }
    return ezxml_parse_fp(fp);
}

char* IC_EndPointSet::get_version(ezxml_t xmlRoot){
    return ezxml_child(xmlRoot, XJDVERSION)->txt;
}

/**
 * decide following class variables.
 *   char* cid;
 *   unsigned mynproc; 
 *   unsigned myao;
 */ 
int IC_EndPointSet::register_program(ezxml_t xmlRoot){
    ezxml_t curComp;
    ezxml_t tmp, tmpstr, cid;
   
    curComp = get_first_comp(xmlRoot);
    
    while(curComp){
        cid = ezxml_child(curComp, COMPID);
        tmp = ezxml_child(curComp, COMPNAME);

        if(cid == NULL){
            fprintf(stderr, "component id is required (Review XJD file)\n");
            break;
        }
        if(tmp == NULL){
            fprintf(stderr, "component name is required (Review XJD file)\n");
            break;
        }
        
        if( !strcmp(this->myprogname, tmp->txt) ){
            this->cid = (char*) strdup(cid->txt);
            
            tmpstr = ezxml_child(curComp, N_NODE); 
            if(tmpstr){
                this->mynproc = (unsigned) atoi(tmpstr->txt);
            }

            tmpstr = ezxml_child(curComp, ARRAYORDER); 
            if(tmpstr){
                if (!strcmp(tmpstr->txt, "IC_COLUMN_MAJOR")){
                    this->myao = IC_EndPoint::IC_COLUMN_MAJOR;
                }
                else if(!strcmp(tmpstr->txt, "IC_ROW_MAJOR")){
                    this->myao = IC_EndPoint::IC_ROW_MAJOR;
                }
                else{ /* error*/
                    fprintf(stderr, "Wrong array order in XJD.\n");
                    break;
                }
            }
            return IC_OK;
        }
        curComp = curComp->next;
    }
    
    return IC_EPSET_INIT_FAILURE;
}


ezxml_t IC_EndPointSet::get_first_comp(ezxml_t xmlRoot){
    ezxml_t plist = ezxml_child(xmlRoot, COMPLIST);
    return ezxml_child(plist, COMP);
}


ezxml_t IC_EndPointSet::get_first_conn(ezxml_t xmlRoot){
    ezxml_t clist = ezxml_child(xmlRoot, CONNLIST);
    return ezxml_child(clist, CONN);
}


char* IC_EndPointSet::get_field_byID(char* id, ezxml_t xmlRoot, char* fieldname){
    ezxml_t curComp;
    ezxml_t cid, tmp;
   
    curComp = get_first_comp(xmlRoot);
    
    while(curComp){
        cid = ezxml_child(curComp, COMPID);
        tmp = ezxml_child(curComp, fieldname);

        if( !strcmp(cid->txt, id) ){
            return tmp->txt;
        }

        curComp = curComp->next;
    }

    return NULL;
}


void IC_EndPointSet::commitArrays(int& stat){
    #define DEBUG
    #define IC_DEBUG
    
    map<string, IC_EndPoint*>::const_iterator iter;
    map<string, IC_Desc*>::const_iterator p2d_iter;
    map<string, IC_Region**>::const_iterator p2r_iter;
    
    for (iter = this->port2ep.begin(); iter != this->port2ep.end(); ++iter) {
        string portname = iter->first;
        IC_EndPoint* cur_ep = (IC_EndPoint*)iter->second;

        IC_Program* myprog = cur_ep->get_myprog();
        IC_Program* otherprog = cur_ep->get_otherprog();
        IC_Desc*    desc = NULL;
        IC_Region** region_set = NULL;

        /* find desc*/
        p2d_iter = this->port2desc.find(portname);
        if(p2d_iter == this->port2desc.end()){
            #ifdef DEBUG
                cout << "no such portname in portname-to-icdesc map. must be 1xN port " << endl;
            #endif
            continue;
        }
        else{
            desc = (IC_Desc*)p2d_iter->second;
        }

        /* find regionset*/
        p2r_iter = this->port2regionset.find(portname);
        if(p2r_iter == this->port2regionset.end()){
            #ifdef DEBUG
                cout << "no such portname in portname-to-regionset map" << endl;
            #endif
            stat = IC_COMMIT_ARRAY_FAILURE;
            return;
        }
        else{
            region_set = (IC_Region**)p2r_iter->second;
        }
        
        assert(myprog != NULL && otherprog != NULL && desc != NULL && region_set != NULL);

        cout << portname << " : " << myprog->name << " " << otherprog->name << endl;

        /* compute schedule for the regionset corresponding to the given portname*/
        /* region set size has been fixed to 1*/
        IC_Sched* sched = IC_Compute_schedule(myprog, otherprog, desc, region_set, 1);
        if (!sched) {
            #ifdef IC_DEBUG
                cout << "port : " << portname << " : "
                     << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] 
                     << endl;
            #endif
            stat = IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
            return;
        }
        else{
            this->port2sched[portname] = sched;
        }
    }
    
    stat = IC_OK;
}


void IC_EndPointSet::exportArray(char* ptname, int& stat){
    #define DEBUG
    
    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, IC_Sched*>::const_iterator p2s_iter; 
    map<string, int>::const_iterator p2tag_iter; 
    
    IC_EndPoint* ep = NULL;
    IC_Program* myprog = NULL;
    IC_Program* otherprog = NULL;
    IC_Sched* sched = NULL;
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        myprog = ep->get_myprog();
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second;
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2s_iter = this->port2sched.find(ptname);
    if(p2s_iter != this->port2sched.end()){
        sched = (IC_Sched*)p2s_iter->second;
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

#ifdef PPLUSPLUS
    map<string, const intArray*>::const_iterator intarr_iter; 
    intarr_iter = this->ppp_int_arrmap.find(ptname);
    if(intarr_iter != this->ppp_int_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in intArrays" << endl;
        #endif

        const intArray* intarr = (const intArray*)intarr_iter->second;

        int ret = IC_Send_int(otherprog, sched, intarr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const floatArray*>::const_iterator floatarr_iter; 
    floatarr_iter = this->ppp_float_arrmap.find(ptname);
    if(floatarr_iter != this->ppp_float_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in floatArrays" << endl;
        #endif

        const floatArray* floatarr = (const floatArray*)floatarr_iter->second;

        int ret = IC_Send_float(otherprog, sched, floatarr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const doubleArray*>::const_iterator doublearr_iter; 
    doublearr_iter = this->ppp_double_arrmap.find(ptname);
    if(doublearr_iter != this->ppp_double_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in doubleArrays" << endl;
        #endif

        const doubleArray* doublearr = (const doubleArray*)doublearr_iter->second;

        int ret = IC_Send_double(otherprog, sched, doublearr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }
#endif

#ifdef FORTRAN90
    map<string, const F90_charArray*>::const_iterator f90chararr_iter; 
    f90chararr_iter = this->f90_char_arrmap.find(ptname);
    if(f90chararr_iter != this->f90_char_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in charArrays" << endl;
        #endif

        const F90_charArray* chararr = (const F90_charArray*)f90chararr_iter->second;

        int ret = IC_Send_char(otherprog, sched, chararr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_shortArray*>::const_iterator f90shortarr_iter; 
    f90shortarr_iter = this->f90_short_arrmap.find(ptname);
    if(f90shortarr_iter != this->f90_short_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in shortArrays" << endl;
        #endif

        const F90_shortArray* shortarr = (const F90_shortArray*)f90shortarr_iter->second;

        int ret = IC_Send_short(otherprog, sched, shortarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_intArray*>::const_iterator f90intarr_iter; 
    f90intarr_iter = this->f90_int_arrmap.find(ptname);
    if(f90intarr_iter != this->f90_int_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in intArrays" << endl;
        #endif

        const F90_intArray* intarr = (const F90_intArray*)f90intarr_iter->second;

        int ret = IC_Send_int(otherprog, sched, intarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_floatArray*>::const_iterator f90floatarr_iter; 
    f90floatarr_iter = this->f90_float_arrmap.find(ptname);
    if(f90floatarr_iter != this->f90_float_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in floatArrays" << endl;
        #endif

        const F90_floatArray* floatarr = (const F90_floatArray*)f90floatarr_iter->second;

        int ret = IC_Send_float(otherprog, sched, floatarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_doubleArray*>::const_iterator f90doublearr_iter; 
    f90doublearr_iter = this->f90_double_arrmap.find(ptname);
    if(f90doublearr_iter != this->f90_double_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in doubleArrays" << endl;
        #endif

        const F90_doubleArray* doublearr = (const F90_doubleArray*)f90doublearr_iter->second;

        int ret = IC_Send_double(otherprog, sched, doublearr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

#endif

    if(found == false){
        #ifdef DEBUG
            cout << "export: no array is registered with the given portname" << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_OK;
}

void IC_EndPointSet::importArray(char* ptname, int& stat){
    #define DEBUG
    
    bool found = false;

    map<string, IC_EndPoint*>::const_iterator p2ep_iter;
    map<string, IC_Sched*>::const_iterator p2s_iter; 
    map<string, int>::const_iterator p2tag_iter; 
    
    IC_EndPoint* ep = NULL;
    IC_Program* myprog = NULL;
    IC_Program* otherprog = NULL;
    IC_Sched* sched = NULL; 
    int tag = -1;

    p2ep_iter = this->port2ep.find(ptname);
    if(p2ep_iter != this->port2ep.end()){
        ep = (IC_EndPoint*)p2ep_iter->second;
        myprog = ep->get_myprog();
        otherprog = ep->get_otherprog();
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2tag_iter = this->port2tag.find(ptname);
    if(p2tag_iter != this->port2tag.end()){
        tag = (int)p2tag_iter->second;
        #ifdef DEBUG
            cout << "port tag found : " << tag << endl;
        #endif
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    p2s_iter = this->port2sched.find(ptname);
    if(p2s_iter != this->port2sched.end()){
        sched = (IC_Sched*)p2s_iter->second;
    }
    else{
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

#ifdef PPLUSPLUS
    map<string, const intArray*>::const_iterator intarr_iter; 
    intarr_iter = this->ppp_int_arrmap.find(ptname);
    if(intarr_iter != this->ppp_int_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in intArrays" << endl;
        #endif

        const intArray* intarr = (const intArray*)intarr_iter->second;

        int ret = IC_Recv_int(otherprog, sched, intarr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const floatArray*>::const_iterator floatarr_iter; 
    floatarr_iter = this->ppp_float_arrmap.find(ptname);
    if(floatarr_iter != this->ppp_float_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in floatArrays" << endl;
        #endif

        const floatArray* floatarr = (const floatArray*)floatarr_iter->second;

        int ret = IC_Recv_float(otherprog, sched, floatarr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const doubleArray*>::const_iterator doublearr_iter; 
    doublearr_iter = this->ppp_double_arrmap.find(ptname);
    if(doublearr_iter != this->ppp_double_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in doubleArrays" << endl;
        #endif

        const doubleArray* doublearr = (const doubleArray*)doublearr_iter->second;

        int ret = IC_Recv_double(otherprog, sched, doublearr->getLocalArray().getDataPointer(), tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }
#endif

#ifdef FORTRAN90
    map<string, const F90_charArray*>::const_iterator f90chararr_iter; 
    f90chararr_iter = this->f90_char_arrmap.find(ptname);
    if(f90chararr_iter != this->f90_char_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in charArrays" << endl;
        #endif

        const F90_charArray* chararr = (const F90_charArray*)f90chararr_iter->second;

        int ret = IC_Recv_char(otherprog, sched, chararr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_shortArray*>::const_iterator f90shortarr_iter; 
    f90shortarr_iter = this->f90_short_arrmap.find(ptname);
    if(f90shortarr_iter != this->f90_short_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in shortArrays" << endl;
        #endif

        const F90_shortArray* shortarr = (const F90_shortArray*)f90shortarr_iter->second;

        int ret = IC_Recv_short(otherprog, sched, shortarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_intArray*>::const_iterator f90intarr_iter; 
    f90intarr_iter = this->f90_int_arrmap.find(ptname);
    if(f90intarr_iter != this->f90_int_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in intArrays" << endl;
        #endif

        const F90_intArray* intarr = (const F90_intArray*)f90intarr_iter->second;

        int ret = IC_Recv_int(otherprog, sched, intarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_floatArray*>::const_iterator f90floatarr_iter; 
    f90floatarr_iter = this->f90_float_arrmap.find(ptname);
    if(f90floatarr_iter != this->f90_float_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in floatArrays" << endl;
        #endif

        const F90_floatArray* floatarr = (const F90_floatArray*)f90floatarr_iter->second;

        int ret = IC_Recv_float(otherprog, sched, floatarr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

    map<string, const F90_doubleArray*>::const_iterator f90doublearr_iter; 
    f90doublearr_iter = this->f90_double_arrmap.find(ptname);
    if(f90doublearr_iter != this->f90_double_arrmap.end()){
        found = true;
        #ifdef DEBUG
            cout << ptname << " is found in doubleArrays" << endl;
        #endif

        const F90_doubleArray* doublearr = (const F90_doubleArray*)f90doublearr_iter->second;

        int ret = IC_Recv_double(otherprog, sched, doublearr->base, tag);
        if (ret!=0) {
            #ifdef IC_DEBUG
                cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
            #endif
            stat = IC_EndPoint::IC_COMM_FAILURE;
            return;
        }
    }

#endif


    if(found == false){
        #ifdef DEBUG
            cout << "import: no array is registered with the given portname" << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }

    stat = IC_OK;
}

void IC_EndPointSet::printErrorMessage(const char* msg, const int ic_err) {
  unsigned nerrs=sizeof(IC_epset_errors)/sizeof(char*);
  if (-ic_err<0 || -ic_err>nerrs) {
    std::cerr << "invalid error code\n";
    return;
  }
  if (msg)
    std::cerr << msg << ": " << IC_epset_errors[-ic_err] << std::endl;
  else
    std::cerr << IC_epset_errors[-ic_err] << std::endl;
}

